/* PLUGIN file to use with plantri.c
 *
 */

//trick for nicer viewing in eclipse
#ifndef MAXN
static int *degree;
static int nv;
static int ne;
static int maxnv;
static int edgebound[2];   /* edge count min,max for polytopes */
static int minpolydeg;
static FILE *msgfile;
static int *facesize;
#define MAXF 0
#define MAXE 0
typedef struct e /* The data type used for edges */
{
  int start; /* vertex where the edge starts */
  int end; /* vertex where the edge ends */
  int rightface; /* face on the right side of the edge
   note: only valid if make_dual() called */
  struct e *prev; /* previous edge in clockwise direction */
  struct e *next; /* next edge in clockwise direction */
  struct e *invers; /* the edge that is inverse to this one */
  struct e *min; /* the least of e and e->invers */
  int mark, index; /* two ints for temporary use;
   Only access mark via the MARK macros. */
  int left_facesize; /* size of the face in prev-direction of the edge.
   Only used for -p option. */
} EDGE;
static EDGE **facestart;
static EDGE **firstedge;
#endif

//#define PRE_FILTER_POLY filter()

//#define PRE_FILTER_SIMPLE filter()

//#define PRE_FILTER_MIN4 triFilter()

//#define FILTER filter2

#define FILTER unfilter

#define PLUGIN_INIT uswitch = TRUE;

#define SUMMARY summary

static unsigned long triangulations = 0;
static unsigned long polytopes = 0;
static unsigned long triRefused = 0;
static unsigned long polRefused = 0;
static unsigned long polFiltered = 0;
static int dual = 0;
static int make_dual();
static void (*_callback)(unsigned char[], unsigned int)=0;
static void compute_dual_code(unsigned char code[]);
static void compute_code(unsigned char code[]);

int plantri_main(int argc, char *argv[]);

static int deg3NodesAreMoreThan(int n)
{
  int count = 0;
  int i;
  for (i = 0; i < nv; ++i)
  {
    if (degree[i] == 3)
    {
      if (++count > n)
      {
        return 1;
      }
    }
  }
  return 0;
}

static int filterWithoutCounting()
{
  //I don't understand what this did and why I had this???
  /*
  if (!((nv <= 1 + 2 * (maxnv - nv)) || !deg3NodesAreMoreThan(1 + 2 * (maxnv - nv))))
  {
    return 0;
  }*/

  //check the following constraints
  //* No large face forks or cycles
  //* Is |E| <= |F| + |T| together with |F| <= n both still possible?


  //check that their are no large face cycles or large face forks
  //we clearly only need to check that if there exist at least 3 large faces
  // |E| == 3 |V| - 6 is equivalent to a triangulation, i.e. no large faces
  // => to have at least 3 large faces we need |E| <= 3 |V| - 9 or in terms of ne (= 2|E|) and nv (=2|V|):
  // ne < 6 nv - 18
  if (ne <= 6 * nv - 18)
  {
    int nf = make_dual();
    unsigned char visited[nf]; //0 ... not visited, 1 ... visited, 2 ... triangle
    int i;
    int nt = 0;
    for (i = 0; i < nf; ++i)
    {
      visited[i] = 0;
      if (facesize[i] == 3) {
        nt++;
      }
    }
    int n_times_2 = edgebound[1] - 2 * maxnv + 4;
    if (ne > 2 * nt + n_times_2) {
      return 0;
    }
    for (i = 0; i < nf; ++i)
    {
      if (facesize[i] > 3)
      {
        int next = i;
        int found = 1;
        int loopCount = 0;
        while (found)
        {
          visited[next] = 1;
          loopCount++;
          found = 0;
          int count = 0;
          EDGE* neighborEdge = facestart[next];
          EDGE* firstEdge = neighborEdge;
          do
          {
            if (neighborEdge->left_facesize > 3)
            {
              int neighborFace = neighborEdge->invers->rightface;
              count++;
              if (count > 2)
              {
                //fork found
                return 0;
              }
              if (visited[neighborFace] == 0)
              {
                found = 1;
                next = neighborFace;
              }
              if (neighborFace == i && loopCount > 2)
              {
                //cycle found
                return 0;
              }
            }
            neighborEdge = neighborEdge->next->invers;
          } while (neighborEdge != firstEdge);
        }
      }
      else
      {
        visited[i] = 2;
      }
    }
  }
  return 1;
}

static int triFilter()
{
  int debug = 1;
  return 1;
}

static int filter()
{
  int res = filterWithoutCounting();
  if (ne == 6 * nv - 12)
  {
    triangulations++;
    if (!res)
    {
      triRefused++;
    }
  }
  polytopes++;
  if (!res)
  {
    polRefused++;
  }
  return res;
}

static int filter2(int nbtot, int nbop, int doflip)
{
  if(!filterWithoutCounting())
  {
    return 0;
  }
  if(minpolydeg <= 3)
  {
    //check if their is a 3deg-vertice with only triangles around
    int i;
    for (i = 0; i < nv; ++i)
    {
      if(degree[i]==3)
      {
        EDGE* e = firstedge[i];
        int found = 0;
        do
        {
          if(e->left_facesize>3)
          {
            found = 1;
            break;
          }
          e = e -> next;
        } while ( e != firstedge[i]);
        //3deg-vertice only has triangles around
        if(!found)
        {
          polFiltered++;
          return 0;
        }
        break;
      }
    }
  }
  //call plantri communicator
  unsigned char code[MAXF+MAXE+1];

  size_t length=3+ne+(ne/2)-nv;
  compute_dual_code(code);
  _callback(code, MAXF + MAXE + 1);
  return 1;
}

static int unfilter(int nbtot, int nbop, int doflip) {
  unsigned char code[MAXF+MAXE+1];
  if (dual == 0) {
    compute_code(code);
  } else {
    compute_dual_code(code);
  }
  _callback(code, MAXF + MAXE + 1);
  return 1;
}

static void summary()
{
  fprintf(msgfile, "Number of triangulations checked: %lu\nNumber of polytopes checked: %lu\n", triangulations,
    polytopes);
  fprintf(msgfile, "Number of triangulations refused: %lu\nNumber of polytopes refused: %lu\n", triRefused, polRefused);
  fprintf(msgfile, "Number of polytopes filtered: %lu\n", polFiltered);
}

void start_plantri(int argc, char *argv[], void (*callback)(unsigned char[], unsigned int), int _dual)
{
  dual = _dual;
  _callback=callback;
  plantri_main(argc, argv);
}
