//
// Created by benedikt on 4/12/19.
//

#ifndef UMC_NEW_GRAPH_H
#define UMC_NEW_GRAPH_H

#include <boost/graph/adjacency_list.hpp>

namespace UMC {
  template<class OutEdgeList = boost::vecS,
      class VertexList = boost::vecS,
      class Directed = boost::directedS,
      class VertexProperties = boost::no_property,
      class EdgeProperties = boost::no_property,
      class GraphProperties = boost::no_property,
      class EdgeList = boost::listS>
  class Graph : public boost::adjacency_list<
      OutEdgeList,
      VertexList,
      Directed,
      VertexProperties,
      boost::property<boost::edge_index_t, uint, EdgeProperties>,
      GraphProperties,
      EdgeList> {
    public:
      typedef uint EdgeId;

      typename boost::graph_traits<Graph>::edge_descriptor get_edge(const EdgeId& edgeId) const {
        return edges.at(edgeId);
      }

      std::pair<typename Graph::edge_descriptor, bool> add_edge(
          typename Graph::vertex_descriptor u,
          typename Graph::vertex_descriptor v,
          const EdgeProperties& p) {
        auto id = (EdgeId) edges.size();
        auto res = boost::add_edge(u, v, typename Graph::edge_property_type(id, p), *this);
        if (res.second) {
          edges.push_back(res.first);
        }
        return res;
      }

      std::pair<typename Graph::edge_descriptor, bool> add_edge(
          typename Graph::vertex_descriptor u,
          typename Graph::vertex_descriptor v) {
        return add_edge(u, v, EdgeProperties());
      }

    private:
      std::vector<typename Graph::edge_descriptor> edges;
  };
}

namespace boost {

  template<class A = boost::vecS,
      class B = boost::vecS,
      class C = boost::directedS,
      class D = boost::no_property,
      class E = boost::no_property,
      class F = boost::no_property,
      class G = boost::listS>
  inline std::pair<typename UMC::Graph<A, B, C, D, E, F, G>::edge_descriptor, bool>
  add_edge(typename UMC::Graph<A, B, C, D, E, F, G>::vertex_descriptor u,
           typename UMC::Graph<A, B, C, D, E, F, G>::vertex_descriptor v,
           const E& p,
           UMC::Graph<A, B, C, D, E, F, G>& g_) {
    return g_.add_edge(u, v, p);
  }

  template<class A = boost::vecS,
      class B = boost::vecS,
      class C = boost::directedS,
      class D = boost::no_property,
      class E = boost::no_property,
      class F = boost::no_property,
      class G = boost::listS>
  inline std::pair<typename UMC::Graph<A, B, C, D, E, F, G>::edge_descriptor, bool>
  add_edge(typename UMC::Graph<A, B, C, D, E, F, G>::vertex_descriptor u,
           typename UMC::Graph<A, B, C, D, E, F, G>::vertex_descriptor v,
           UMC::Graph<A, B, C, D, E, F, G>& g_) {
    E p;
    return add_edge(u, v, p, g_);
  }
}

#endif //UMC_NEW_GRAPH_H
