//
// Created by benedikt on 4/16/19.
//

#ifndef UMC_NEW_OUTPUTSTATISTICS_H
#define UMC_NEW_OUTPUTSTATISTICS_H

namespace UMC {
  struct OutputStatistics {
      double runtime = 0.0; // in seconds
      unsigned long ilps_solved = 0;
      unsigned long max_flows_calculated = 0;
      unsigned long cycles_found = 0;
      unsigned long cycles_inserted = 0;
      unsigned long cycles_removed = 0;
      unsigned long cut_constraints_added = 0;
      unsigned long bugs_occurred = 0;
      unsigned long max_num_vertex_complements = 0;

      void reset(bool reset_bugs = true) {
        runtime = 0.0;
        ilps_solved = 0;
        max_flows_calculated = 0;
        cycles_found = 0;
        cut_constraints_added = 0;
        cycles_removed = 0;
        max_num_vertex_complements = 0;
        cycles_inserted = 0;
        if (reset_bugs) bugs_occurred = 0;
      }

      void print_compact(std::ostream& out) const {
        out << runtime << "," << ilps_solved << "," << max_flows_calculated << "," << cycles_found << "," <<
            cycles_inserted << "," << cycles_removed << "," << cut_constraints_added << "," << bugs_occurred << ","
            << max_num_vertex_complements;
      }
  };

  std::ostream& operator<<(std::ostream& out, const OutputStatistics& s) {
    out << "Runtime: " << s.runtime << std::endl;
    out << "ILPs solved: " << s.ilps_solved << std::endl;
    out << "Max-Flows calculated: " << s.max_flows_calculated << std::endl;
    out << "Cycles found: " << s.cycles_found << std::endl;
    out << "Cycles inserted: " << s.cycles_inserted << std::endl;
    out << "Cycles removed: " << s.cycles_removed << std::endl;
    out << "Cut Constraints added: " << s.cut_constraints_added << std::endl;
    out << "Bugs occurred: " << s.bugs_occurred << std::endl;
    out << "Maximum number of vertex complements: " << s.max_num_vertex_complements << std::endl;
    return out;
  }
}


#endif //UMC_NEW_OUTPUTSTATISTICS_H
