//
// Created by benedikt on 6/13/19.
//

#ifndef UMC_NEW_K_CONNECTED_H
#define UMC_NEW_K_CONNECTED_H

#include "../Graph.h"
#include "min_max_degree.h"
#include <algorithm>
#include <stack>
#include <tuple>
#include <boost/graph/copy.hpp>

namespace boost {
  //the following method only works for 3-connected graphs
  template<class UndirectedGraph>
  bool is_essentially_k_edge_connected_naive(const UndirectedGraph& g, size_t k) {
    typedef boost::adjacency_list_traits<boost::vecS, boost::vecS, boost::directedS> Traits;
    typedef UMC::Graph<boost::vecS, boost::vecS, boost::directedS,
        boost::no_property,
        boost::property<boost::edge_capacity_t, long,
            boost::property<boost::edge_residual_capacity_t, long,
                boost::property<boost::edge_reverse_t, Traits::edge_descriptor> > >
    > DirectedGraph;
    DirectedGraph g_dir;
    boost::property_map<DirectedGraph, boost::edge_reverse_t>::type rev = get(boost::edge_reverse, g_dir);
    //boost::property_map<DirectedGraph, boost::edge_residual_capacity_t>::type residual_capacity =
    //    get(boost::edge_residual_capacity, g_dir);

    auto n = boost::num_vertices(g);
    auto m = boost::num_edges(g);
    typedef typename boost::graph_traits<DirectedGraph>::vertex_descriptor vertex_descriptor;
    typedef typename boost::graph_traits<DirectedGraph>::edge_descriptor edge_descriptor;
    vertex_descriptor vs[255]; //list of vertices, only n <= 255 are really used
    for (size_t i = 0; i < n; i++) {
      vs[i] = add_vertex(g_dir);
    }
    for (int i = 0; i < 4; i++) { //i == 0: undireced edges i == 1: directed duplicates, i == 2: reverse of
      // undirected edge, i == 3: reverse of directed duplicate
      for (size_t id = 0; id < m; id++) {
        auto e = g.get_edge(id);
        auto v1 = boost::source(e, g);
        auto v2 = boost::target(e, g);
        edge_descriptor e_dir;
        bool in;
        boost::tie(e_dir, in) = g_dir.add_edge(vs[(i == 0 || i == 3) ? v1 : v2], vs[(i == 0 || i == 3) ? v2 : v1]);
        assert(in);
        if (i >= 2) {
          rev[e_dir] = g_dir.get_edge(id + (i == 3 ? m : 0));
          rev[rev[e_dir]] = e_dir;
        }
      }
    }

    auto edge_index = get(boost::edge_index, g_dir);

    for (auto s : boost::make_iterator_range(vertices(g))) {
      for (auto t: boost::make_iterator_range(vertices(g))) {
        if (s < t && !boost::edge(s, t, g).second) {
          //there exists no edge between source and target we can compute the max flow between them
          //edges incident to source or target have capacity k all others have capacity k-1
          boost::property_map<DirectedGraph, boost::edge_capacity_t>::type capacity = get(boost::edge_capacity,
                                                                                          g_dir);
          boost::property_map<DirectedGraph, boost::edge_residual_capacity_t>::type residual_capacity =
              get(boost::edge_residual_capacity, g_dir);

          auto ds = std::min(boost::out_degree(s, g), k);
          auto dt = std::min(boost::out_degree(t, g), k);
          auto def = ds * dt;
          for (const auto& e : boost::make_iterator_range(boost::edges(g_dir))) {
            auto id = get(edge_index, e);
            if (id < 2 * m) {
              if (source(e, g_dir) == s) {
                if (target(e, g_dir) == t) {
                  capacity[e] = std::min(ds, dt) * k;
                } else {
                  capacity[e] = k * dt;
                }
              } else if (target(e, g_dir) == t) {
                capacity[e] = k * ds;
              } else {
                capacity[e] = def;
              }
            } else {
              capacity[e] = 0;
            }
            residual_capacity[e] = capacity[e];
          }

          auto flow = boost::push_relabel_max_flow(g_dir, s, t);
          if (flow < k * def) {
            return false;
          }
        }
      }
    }
    return true;
  }


  namespace internal_3_edge_connected {
    template<class UndirectedGraph>
    struct data {
        typedef typename graph_traits<UndirectedGraph>::vertex_descriptor Vertex;
        typedef typename graph_traits<UndirectedGraph>::edge_descriptor Edge;
        typedef std::tuple<Edge, Vertex, Vertex> StackElement;

        bool find_cut_pairs(Vertex v) {
          visited[v] = true;
          dfs_[v] = dfs;
          dfs++;
          nd[v] = 1;
          lowpt[v] = dfs_[v];
          low[v] = v;
          lowpt2nd[v] = dfs_[v];
          low2nd[v] = v;

          auto edge_index = get(boost::edge_index, *g);

          for (auto e: boost::make_iterator_range(boost::out_edges(v, *g))) { //1.
            auto w = boost::target(e, *g);
            auto e_id = get(edge_index, e);
            if (!visited[w]) { //1.1.
              parent[w] = v;
              tree_edges[e_id] = true;
              if (find_cut_pairs(w)) {
                return true;
              }
              if (!stack[w].empty() && w == std::get<2>(stack[w].top())) { // 1.1.1
                //found cut pair
                return true;
              }
              nd[v] = nd[v] + nd[w];
              if (lowpt[w] < lowpt[v]) { //1.1.2
                lowpt2nd[v] = lowpt[v];
                lowpt[v] = lowpt[w];
                low2nd[v] = low[v];
                low[v] = low[w];
                stack[v] = std::move(stack[w]);
                to_low[v] = e;
              } else if (lowpt[w] < lowpt2nd[v]) { //1.1.3
                lowpt2nd[v] = lowpt[w];
                low2nd[v] = low[w];
                stack[w] = std::stack<StackElement>();
              }
            } else if (dfs_[v] > dfs_[w] && !tree_edges[e_id]) { //1.2.
              if (dfs_[w] <= lowpt[v]) { //1.2.1
                lowpt2nd[v] = lowpt[v];
                lowpt[v] = dfs_[w];
                low2nd[v] = low[v];
                low[v] = w;
                stack[v] = std::stack<StackElement>();
                to_low[v] = e;
              } else if (dfs_[w] < lowpt2nd[v]) {
                lowpt2nd[v] = dfs_[w];
                low2nd[v] = w;
              }
            }
          }
          //2.
          if (stack[v].empty()) { //2.1
            if (lowpt2nd[v] > lowpt[v]) { //2.1.1
              stack[v].push({to_low[v], low[v], low2nd[v]});
            }
          } else {
            if (lowpt2nd[v] > dfs_[std::get<2>(stack[v].top())]) { // 2.2
              stack[v].push({to_low[v], std::get<2>(stack[v].top()), low2nd[v]}); //2.2.1
            } else {
              while (!stack[v].empty() && lowpt2nd[v] <= dfs_[std::get<1>(stack[v].top())]) { //2.2.2
                stack[v].pop();
              }
              if (!stack[v].empty() && lowpt2nd[v] < dfs_[std::get<2>(stack[v].top())]) {
                std::get<2>(stack[v].top()) = low2nd[v];
              }
            }
          }
          //3.
          for (auto e: boost::make_iterator_range(boost::out_edges(v, *g))) {
            auto e_id = get(edge_index, e);
            auto u = boost::target(e, *g);
            assert(visited[u]);
            if (dfs_[u] > dfs_[v] && !tree_edges[e_id]) {
              while (!stack[v].empty() && tree_edges[get(edge_index, std::get<0>(stack[v].top()))] &&
                     dfs_[boost::target(std::get<0>(stack[v].top()), *g)] <= dfs_[u] &&
                     dfs_[u] <= dfs_[boost::target(std::get<0>(stack[v].top()), *g)] +
                                nd[boost::target(std::get<0>(stack[v].top()), *g)] - 1) {
                stack[v].pop();
              }
            }
          }
          return false;
        }

        size_t dfs;
        std::vector<size_t> dfs_;
        std::vector<size_t> nd;
        std::vector<size_t> lowpt;
        std::vector<Vertex> low;
        std::vector<size_t> lowpt2nd;
        std::vector<Vertex> low2nd;
        std::vector<bool> visited;
        const UndirectedGraph* g;
        std::vector<std::stack<StackElement>> stack;
        std::vector<Edge> to_low;
        std::vector<Vertex> parent;
        std::vector<bool> tree_edges;
    };
  }


  // linear time algorithm by Yung H.Tsin see https://www.sciencedirect.com/science/article/pii/S1570866708000415
  template<class UndirectedGraph>
  bool is_3_edge_connected(const UndirectedGraph& g,
                           typename internal_3_edge_connected::data<UndirectedGraph>::Vertex root = 0) {
    //init everything
    typedef typename internal_3_edge_connected::data<UndirectedGraph>::Vertex Vertex;
    typedef typename internal_3_edge_connected::data<UndirectedGraph>::Edge Edge;
    typedef typename internal_3_edge_connected::data<UndirectedGraph>::StackElement StackElement;
    auto n = boost::num_vertices(g);
    internal_3_edge_connected::data<UndirectedGraph> data{
        1,
        std::vector<size_t>(n),
        std::vector<size_t>(n),
        std::vector<size_t>(n),
        std::vector<Vertex>(n),
        std::vector<size_t>(n),
        std::vector<Vertex>(n),
        std::vector<bool>(n, false),
        &g,
        std::vector<std::stack<StackElement>>(n),
        std::vector<Edge>(n),
        std::vector<Vertex>(n),
        std::vector<bool>(boost::num_edges(g), false)
    };
    return !data.find_cut_pairs(root);
  }

  namespace internal_essentially_4_edge_connected {
    struct do_nothing {
        template<typename VertexOrEdge1, typename VertexOrEdge2>
        void operator()(const VertexOrEdge1&, VertexOrEdge2&) const {
        }
    };
  }

  // assumes that g is 3-connected and triangle free
  template<class UndirectedGraph>
  bool is_essentially_4_edge_connected(const UndirectedGraph& g) {
    assert(is_3_edge_connected(g));
    //for each edge remove it, replace 2-degree vertices by edges and check for 3-edge-connectedness.
    typedef boost::adjacency_list<boost::vecS, boost::vecS, boost::undirectedS, no_property,
        property<edge_index_t, std::size_t>> SimpleUndirected;
    SimpleUndirected copy;
    boost::copy_graph(g, copy,
                      boost::vertex_copy(internal_essentially_4_edge_connected::do_nothing())
                          .edge_copy(internal_essentially_4_edge_connected::do_nothing()));
    for (size_t i = 0; i < num_vertices(g) - 3; i++) {
      auto ed = *edges(copy).first;
      auto v = boost::source(ed, copy);
      auto w = boost::target(ed, copy);

      auto dv = boost::out_degree(v, copy);
      auto dw = boost::out_degree(w, copy);
      if (dv < dw) {
        auto tmp = v;
        v = w;
        w = tmp;
        auto tmp2 = dv;
        dv = dw;
        dw = tmp2;
      }

      //std::cout << "before removing: " << std::endl;
      //boost::print_graph(copy);

      boost::remove_edge(ed, copy);

      typedef typename graph_traits<SimpleUndirected>::vertex_descriptor Vertex;
      typedef typename graph_traits<SimpleUndirected>::edge_descriptor Edge;

      Vertex nv1 = -1;
      Vertex nv2 = -1;
      Vertex nw1 = -1;
      Vertex nw2 = -1;
      Edge ev;
      Edge ew;

      if (dw == 3) {
        //clear vertex v and add edge between remaining neigbhors
        auto[it, it_end] = boost::adjacent_vertices(w, copy);
        nw1 = *it++;
        nw2 = *it++;
        assert(it == it_end);
        boost::clear_vertex(w, copy);
        if (nw1 != v || nw2 != v) ew = boost::add_edge(nw1, nw2, copy).first;

        if (dv == 3) {
          //clear vertex v and add edge between remaining neigbhors
          auto[it, it_end] = boost::adjacent_vertices(v, copy);
          nv1 = *it++;
          nv2 = *it++;
          assert(it == it_end);
          boost::clear_vertex(v, copy);
          ev = boost::add_edge(nv1, nv2, copy).first;
        }
      }

      //assign ids to edges
      auto edge_index = get(boost::edge_index, copy);
      size_t id = 0;
      for (auto e: boost::make_iterator_range(boost::edges(copy))) {
        edge_index[e] = id++;
      }

      //std::cout << "before merging: " << std::endl;
      //boost::print_graph(copy);
      if ((nw1 != v || nw2 != v) && !is_3_edge_connected(copy, dv == 3 ? nv1 : v)) {
        return false;
      }

      //undo everything and merge edge ed
      if (dw == 3) {
        if (nw1 != v || nw2 != v) boost::remove_edge(ew, copy);
        if (nw1 != v) boost::add_edge(nw1, v, copy);
        if (nw2 != v) boost::add_edge(nw2, v, copy);
        if (dv == 3) {
          boost::remove_edge(ev, copy);
          boost::add_edge(nv1, v, copy);
          boost::add_edge(nv2, v, copy);
        }
      } else {
        boost::remove_edge(v, w, copy); //there may be more edges between v, w => those would be loops after merging
        // => remove them
        for (auto e: boost::make_iterator_range(boost::out_edges(w, copy))) {
          auto x = target(e, copy);
          assert(x != w && x != v);
          boost::add_edge(v, x, copy);
        }
        boost::clear_vertex(w, copy);
      }
      //std::cout << "after merging: " << std::endl;
      //boost::print_graph(copy);
      if (boost::out_degree(v, copy) <= 3) {
        assert(boost::out_degree(v, copy) == 3);
        return false;
      }
    }
    return true;
  }
}

#endif //UMC_NEW_K_CONNECTED_H