//
// Created by benedikt on 6/27/19.
//

#ifndef UMC_NEW_PARSE_GRAPHS_H
#define UMC_NEW_PARSE_GRAPHS_H

#include <istream>
#include <vector>
#include <string>

namespace boost {

  namespace graph6_internal {
    std::vector<bool> parse_R(std::istream& in, std::size_t size) {
      std::vector<bool> res(size, false);
      auto byte_length = (size + 5) / 6;
      for (size_t i = 0; i < byte_length; i++) {
        char c = in.get();
        assert(c >= 63);
        c -= 63;
        for (size_t j = 5; j != (size_t) -1; j--) {
          auto val = c % 2;
          if (val == 1) {
            assert(6 * i + j < size);
            res[6 * i + j] = true;
          }
          c /= 2;
        }
        assert(c == 0);
      }
      return res;
    }

    template<class T>
    std::vector<T> parse_r(std::istream& in, std::size_t size, T max) {
      std::vector<T> res;
      res.reserve(size);
      T bitPerEl, var = max;
      for (bitPerEl = 0; var != 0; ++bitPerEl) var >>= 1;

      std::vector<bool> bits = parse_R(in, bitPerEl * size);

      for (size_t i = 0; i < size; i++) {
        T val = 0;
        for (T k = 0; k < bitPerEl; k++) {
          if (bits[i * bitPerEl + k]) {
            val += 1;
          }
          if (k < bitPerEl - 1) {
            val <<= 1;
          }
        }
        assert(val <= max);
        res.push_back(val);
      }
      return res;
    }

    unsigned long parse_N(std::istream& in) {
      char c = in.get();
      if (c != 126) {
        assert (c < 126 && c >= 63);
        return c - 63;
      }
      c = in.peek();
      size_t size = c != 126 ? 18 : 36;
      if (c == 126) {
        in.get();
      }
      auto binary = parse_R(in, size);
      assert(binary.size() == size);
      std::bitset<36> bits;
      for (size_t i = 0; i < size; i++) {
        if (binary[size - 1 - i]) {
          bits[i] = true;
        }
      }
      return bits.to_ulong();
    }
  }

  template<class UndirectedGraph>
  void parse_graph6(std::istream& in, UndirectedGraph& g) {
    auto n = graph6_internal::parse_N(in);
    for (size_t i = 0; i < n; i++) {
      boost::add_vertex(g);
    }

    std::vector<bool> x = graph6_internal::parse_R(in, (n * (n - 1)) / 2);
    size_t index = 0;
    for (size_t j = 1; j < n; j++) {
      for (size_t i = 0; i < j; i++) {
        if (x[index]) {
          boost::add_edge(i, j, g);
        }
        index++;
      }
    }
  }

  template<class UndirectedGraph>
  void parse_osparse6(std::istream& in, UndirectedGraph& g) {
    auto c = in.get();
    assert(c == '0');
    _unused(c);
    auto n = graph6_internal::parse_N(in);
    for (size_t i = 0; i < n; i++) {
      boost::add_vertex(g);
    }
    auto m = graph6_internal::parse_N(in);
    std::vector<size_t> x = graph6_internal::parse_r(in, 2 * m, n - 1);
    assert(x.size() == 2 * m);
    for (size_t i = 0; i < m; i++) {
      auto source = x[2 * i];
      auto target = x[2 * i + 1];
      boost::add_edge(source, target, g);
    }
    assert(boost::num_edges(g) == m);
  }

  /**
   * parses one of the graph6 formats depending on the first character
   * @tparam UndirectedGraph
   * @param in
   * @param g
   */
  template<class UndirectedGraph>
  void parse6(std::istream& in, UndirectedGraph& g) {
    auto c = in.peek();
    switch (c) {
    case '0' :
      parse_osparse6(in, g);
      return;
    default:
      parse_graph6(in, g);
      return;
    }
  }
}

#endif //UMC_NEW_PARSE_GRAPHS_H
