//
// Created by benedikt on 6/11/19.
//

#ifndef UMC_NEW_PLANAR_DUAL_H
#define UMC_NEW_PLANAR_DUAL_H

#include <boost/graph/planar_face_traversal.hpp>
#include <boost/graph/boyer_myrvold_planar_test.hpp>

namespace boost {

  namespace planar_dual_visitor {
    template<class Graph, class MutableGraph>
    struct visitor : public planar_face_traversal_visitor {
        void begin_face() {
          current_face = add_vertex(*graph);
        }

        template<typename Edge>
        void next_edge(Edge e) {
          auto index = get(*edge_index_map, e);
          if (edge_to_face.count(index) > 0) {
            add_edge(edge_to_face.at(index), current_face, *graph);
          } else {
            edge_to_face[index] = current_face;
          }
        }


        typedef typename property_map<Graph, edge_index_t>::const_type IndexMap;
        typedef typename property_traits<IndexMap>::value_type EdgeIndex;
        std::unordered_map<EdgeIndex, typename graph_traits<MutableGraph>::vertex_descriptor> edge_to_face;
        typename graph_traits<MutableGraph>::vertex_descriptor current_face;
        MutableGraph* graph;
        IndexMap* edge_index_map;
    };
  }

  template<class Graph, class MutableGraph>
  void planar_dual(Graph const& g, MutableGraph& dual,
                   const std::vector<std::vector<typename graph_traits<Graph>::edge_descriptor>>* embedding = nullptr) {
    typedef std::vector<typename graph_traits<Graph>::edge_descriptor> vec_t;
    std::vector<vec_t> _embedding(num_vertices(g));
    if (embedding == nullptr) {
      bool planar = boyer_myrvold_planarity_test(boyer_myrvold_params::graph = g,
                                                 boyer_myrvold_params::embedding = &_embedding.at(0));
      assert(planar);
      _unused(planar);
      embedding = &_embedding;
    }

    planar_dual_visitor::visitor<Graph, MutableGraph> v;
    v.graph = &dual;
    typename boost::property_map<Graph, edge_index_t>::const_type edge_index_map = get(edge_index, g);
    v.edge_index_map = &edge_index_map;
    BOOST_CONCEPT_ASSERT((VertexAndEdgeListGraphConcept<Graph>));
    planar_face_traversal(g, &(*embedding).at(0), v);
  }
}

#endif //UMC_NEW_PLANAR_DUAL_H
