//
// Created by benedikt on 6/25/19.
//

#ifndef UMC_NEW_PRINTING_GRAPHS_H
#define UMC_NEW_PRINTING_GRAPHS_H

#include <ostream>
#include <vector>
#include <string>

namespace boost {

  namespace graph6_internal {
    void R(std::ostream& out, const std::vector<bool>& x) {
      for (size_t i = 0; i < x.size(); i += 6) {
        char v = 0;
        for (size_t j = 0; j < 6; j++) {
          if (j > 0) {
            v *= 2;
          }
          if (i + j < x.size() && x[i + j]) {
            v++;
          }
        }
        out << (char)(v+63);
      }
    }

    template<class T>
    void r(std::ostream& out, const std::vector<T>& x, T max) {
      T bitPerEl, var = max;
      for (bitPerEl = 0; var != 0; ++bitPerEl) var >>= 1;

      std::vector<bool> bits(bitPerEl * x.size(), false);
      for (size_t i = 0; i < x.size(); i++) {
        auto val = x[i];
        assert(val <= max);
        for (size_t k = bitPerEl - 1; k != (size_t) -1; k--) {
          if (val % 2 == 1) {
            bits[i * bitPerEl + k] = true;
          }
          val >>= 1;
        }
        assert(val == 0);
      }
      R(out, bits);
    }

    void N(std::ostream& out, unsigned long n) {
      if (n <= 62) {
        out << (char)(n + 63);
      } else {
        assert(n < ((unsigned long) 1 << 36));
        int size = n <= 258047 ? 18 : 36;
        std::bitset<36> bits(n);
        std::vector<bool> x;
        x.reserve(size);
        for (size_t i = size - 1; i != (size_t) -1; i--) {
          x.push_back(bits[i]);
        }
        out << (char) 126;
        if (size == 36) {
          out << (char) 126;
        }
        R(out, x);
      }
    }
  }

  template<class UndirectedGraph>
  void print_graph6(std::ostream& out, const UndirectedGraph& g) {
    auto n = boost::num_vertices(g);
    graph6_internal::N(out, n);
    std::vector<bool> x;
    x.reserve((n * (n - 1)) / 2);
    for (size_t j = 1; j < n; j++) {
      for (size_t i = 0; i < j; i++) {
        x.push_back(boost::edge(i, j, g).second);
      }
    }
    graph6_internal::R(out, x);
  }

  /**
   * Ordered version of sparse6 graph
   * Starts with 0 to distinguish from other formats
   * @tparam UndirectedGraph
   * @param out
   * @param g
   */
  template<class UndirectedGraph>
  void print_osparse6(std::ostream& out, const UndirectedGraph& g) {
    auto n = boost::num_vertices(g);
    out << "0";
    graph6_internal::N(out, n);
    auto m = boost::num_edges(g);
    graph6_internal::N(out, m);
    std::vector<size_t> x;
    x.reserve(2 * m);
    for (size_t i = 0; i < m; i++) {
      auto e = g.get_edge(i);
      auto source = boost::source(e, g);
      auto target = boost::target(e, g);
      x.push_back(source);
      x.push_back(target);
    }
    graph6_internal::r(out, x, n - 1);
  }

}


#endif //UMC_NEW_PRINTING_GRAPHS_H
