//
// Created by benedikt on 4/11/19.
//

#ifndef UMC_NEW_PLANAR_CODE_H
#define UMC_NEW_PLANAR_CODE_H

#include <cstring>
#include <istream>
#include <cassert>
#include <boost/graph/graph_traits.hpp>
#include <boost/interprocess/streams/bufferstream.hpp>
#include <boost/graph/graph_utility.hpp>
#include "global.h"

namespace UMC {
  /*struct membuf : std::streambuf {
      membuf(char* begin, char* end) {
        this->setg(begin, begin, end);
      }
  };*/

/**
 * Reads the next graph given in the input stream in planar code format
 * (see https://users.cecs.anu.edu.au/~bdm/plantri/plantri-guide.txt for a format definition)
 * @tparam Graph the boost graph class
 * @param in the in stream
 * @param graph the boost graph to fill
 * @return true if a graph could be read
 */
  template<typename Graph>
  bool read_planar_code(std::istream& in, Graph& graph) {
    typedef typename boost::graph_traits<Graph>::vertex_descriptor vertex_descriptor;
    int n = in.get(); //number of vertices in the graph
    if (n == '>' && in.peek() == '>') {
      //input starts with >>planar_code<<
      char start[14];
      in.get(start, 14);
      assert(std::strcmp(start, ">planar_code<<") == 0);
      n = in.get();
    }
    if (n == EOF) {
      return false;
    }

    vertex_descriptor vs[255]; //list of vertices, only n <= 255 are really used

    for (int i = 0; i < n; i++) {
      vs[i] = add_vertex(graph);
    }

    for (int i = 0; i < n; i++) {
      int c;
      while ((c = in.get()) != 0) {
        c--;
        if (c > i) {
          //add edge between i and c
          auto v = vs[i];
          if (UMC::internal::d2prob > 0.0) {
            if (UMC::internal::udis(UMC::internal::gen) < UMC::internal::d2prob) {
              v = add_vertex(graph);
              graph.add_edge(vs[i], v);
            }
          }
          graph.add_edge(v, vs[c]);
        }
      }
    }

    return true;
  }

  template<typename Graph>
  bool parse_planar_code(char code[], unsigned int maxN, Graph& graph) {
    boost::interprocess::bufferstream in(code, maxN);
    return read_planar_code(in, graph);
  }
}


#endif //UMC_NEW_PLANAR_CODE_H
